import os
from typing import Optional

import numpy as np
import pvkoala
from fastapi import WebSocket, WebSocketDisconnect
from loguru import logger

from vocode.streaming.agent.abstract_factory import AbstractAgentFactory
from vocode.streaming.models.agent import AgentConfig
from vocode.streaming.models.events import PhoneCallConnectedEvent
from vocode.streaming.models.synthesizer import SynthesizerConfig
from vocode.streaming.models.telephony import PhoneCallDirection, VonageConfig
from vocode.streaming.models.transcriber import TranscriberConfig
from vocode.streaming.output_device.vonage_output_device import VonageOutputDevice
from vocode.streaming.synthesizer.abstract_factory import AbstractSynthesizerFactory
from vocode.streaming.telephony.client.vonage_client import VonageClient
from vocode.streaming.telephony.config_manager.base_config_manager import BaseConfigManager
from vocode.streaming.telephony.conversation.abstract_phone_conversation import (
    AbstractPhoneConversation,
)
from vocode.streaming.transcriber.abstract_factory import AbstractTranscriberFactory
from vocode.streaming.utils.events_manager import EventsManager
from vocode.streaming.utils.state_manager import VonagePhoneConversationStateManager

KOALA_CHUNK_SIZE = 512  # 16 bit samples, size 256


class VonagePhoneConversation(AbstractPhoneConversation[VonageOutputDevice]):
    telephony_provider = "vonage"

    def __init__(
        self,
        direction: PhoneCallDirection,
        from_phone: str,
        to_phone: str,
        base_url: str,
        config_manager: BaseConfigManager,
        agent_config: AgentConfig,
        transcriber_config: TranscriberConfig,
        synthesizer_config: SynthesizerConfig,
        vonage_uuid: str,
        vonage_config: VonageConfig,
        agent_factory: AbstractAgentFactory,
        transcriber_factory: AbstractTranscriberFactory,
        synthesizer_factory: AbstractSynthesizerFactory,
        conversation_id: Optional[str] = None,
        events_manager: Optional[EventsManager] = None,
        output_to_speaker: bool = False,
        speed_coefficient: float = 1.0,
        per_chunk_allowance_seconds: float = 0.01,
        noise_suppression: bool = False,
    ):
        self.speed_coefficient = speed_coefficient
        super().__init__(
            direction=direction,
            speed_coefficient=speed_coefficient,
            from_phone=from_phone,
            to_phone=to_phone,
            base_url=base_url,
            config_manager=config_manager,
            output_device=VonageOutputDevice(output_to_speaker=output_to_speaker),
            agent_config=agent_config,
            transcriber_config=transcriber_config,
            synthesizer_config=synthesizer_config,
            conversation_id=conversation_id,
            events_manager=events_manager,
            transcriber_factory=transcriber_factory,
            agent_factory=agent_factory,
            synthesizer_factory=synthesizer_factory,
            per_chunk_allowance_seconds=per_chunk_allowance_seconds,
        )
        self.vonage_config = vonage_config
        self.telephony_client = VonageClient(
            base_url=self.base_url,
            maybe_vonage_config=self.vonage_config,
        )
        self.vonage_uuid = vonage_uuid
        self.noise_suppression = noise_suppression
        if self.noise_suppression:
            logger.info("Using PV koala noise suppression")
            self.buffer = bytearray()
            self.koala = pvkoala.create(
                access_key=os.environ["KOALA_ACCESS_KEY"],
            )

    def create_state_manager(self) -> VonagePhoneConversationStateManager:
        return VonagePhoneConversationStateManager(self)

    async def attach_ws_and_start(self, ws: WebSocket):
        # start message
        await ws.receive()
        super().attach_ws(ws)

        await self.start()
        self.events_manager.publish_event(
            PhoneCallConnectedEvent(
                conversation_id=self.id,
                to_phone_number=self.to_phone,
                from_phone_number=self.from_phone,
            )
        )
        disconnected = False
        while self.active:
            try:
                message = await ws.receive()
                if message["type"] == "websocket.disconnect":
                    raise WebSocketDisconnect(message["code"])
                if "bytes" in message:
                    chunk = message["bytes"]
                    self.receive_audio(chunk)
                else:
                    logger.debug(f"Received non-bytes message: {message}")
            except WebSocketDisconnect as e:
                logger.debug("Websocket disconnected")
                if e.code != 1000:
                    logger.error(f"Websocket disconnected abnormally with code {e.code} {e.reason}")
                disconnected = True
                break
        await self.terminate()
        if not disconnected:
            await ws.close()

    def receive_audio(self, chunk: bytes):
        if self.noise_suppression:
            self.buffer.extend(chunk)

            while len(self.buffer) >= KOALA_CHUNK_SIZE:
                koala_chunk = np.frombuffer(self.buffer[:KOALA_CHUNK_SIZE], dtype=np.int16)
                try:
                    denoised_chunk = np.array(
                        self.koala.process(koala_chunk), dtype=np.int16
                    ).tobytes()
                except Exception:
                    denoised_chunk = koala_chunk.tobytes()
                super().receive_audio(denoised_chunk)
                self.buffer = self.buffer[KOALA_CHUNK_SIZE:]
        else:
            super().receive_audio(chunk)



from typing import Optional

from vocode.streaming.models.telephony import BaseCallConfig


class BaseConfigManager:
    async def save_config(self, conversation_id: str, config: BaseCallConfig):
        raise NotImplementedError

    async def get_config(self, conversation_id) -> Optional[BaseCallConfig]:
        raise NotImplementedError

    async def delete_config(self, conversation_id):
        raise NotImplementedError



from enum import Enum
from typing import List, Literal, Optional, Union

from pydantic.v1 import validator

from .model import BaseModel, TypedModel
from .vector_db import VectorDBConfig
from vocode.streaming.models.actions import ActionConfig
from vocode.streaming.models.message import BaseMessage

FILLER_AUDIO_DEFAULT_SILENCE_THRESHOLD_SECONDS = 0.5
LLM_AGENT_DEFAULT_TEMPERATURE = 1.0
LLM_AGENT_DEFAULT_MAX_TOKENS = 256
LLM_AGENT_DEFAULT_MODEL_NAME = "text-curie-001"
CHAT_GPT_AGENT_DEFAULT_MODEL_NAME = "gpt-3.5-turbo-1106"
CHAT_GPT_AGENT_16K_MODEL_NAME = "gpt-3.5-turbo-0613-16k"
ACTION_AGENT_DEFAULT_MODEL_NAME = "gpt-3.5-turbo-0613"
CHAT_ANTHROPIC_DEFAULT_MODEL_NAME = "claude-3-haiku-20240307"
CHAT_VERTEX_AI_DEFAULT_MODEL_NAME = "chat-bison@001"
AZURE_OPENAI_DEFAULT_API_TYPE = "azure"
AZURE_OPENAI_DEFAULT_API_VERSION = "2023-07-01-preview"
AZURE_OPENAI_DEFAULT_ENGINE = "gpt-35-turbo"
AZURE_OPENAI_GPT_35_16K_ENGINE = "gpt-35-turbo-16k"
AZURE_OPENAI_GPT_4_ENGINE = "vocode-api-gpt4"
OPENAI_GPT_4_MODEL_NAME = "gpt-4"
OPENAI_GPT_4_32K_MODEL_NAME = "gpt-4-32k"
OPENAI_GPT_4_O_MODEL_NAME = "gpt-4o"
OPENAI_GPT_35_TURBO_1106_MODEL_NAME = "gpt-3.5-turbo-1106"
OPENAI_GPT_4_1106_PREVIEW_MODEL_NAME = "gpt-4-1106-preview"
ANTHROPIC_CLAUDE_3_HAIKU_MODEL_NAME = "claude-3-haiku-20240307"
ANTHROPIC_CLAUDE_3_SONNET_MODEL_NAME = "claude-3-sonnet-20240229"
ANTHROPIC_CLAUDE_3_OPUS_MODEL_NAME = "claude-3-opus-20240229"

InterruptSensitivity = Literal["low", "high"]


class AgentType(str, Enum):
    BASE = "agent_base"
    LLM = "agent_llm"
    CHAT_GPT_ALPHA = "agent_chat_gpt_alpha"
    CHAT_GPT = "agent_chat_gpt"
    ANTHROPIC = "agent_anthropic"
    CHAT_VERTEX_AI = "agent_chat_vertex_ai"
    ECHO = "agent_echo"
    GPT4ALL = "agent_gpt4all"
    LLAMACPP = "agent_llamacpp"
    INFORMATION_RETRIEVAL = "agent_information_retrieval"
    RESTFUL_USER_IMPLEMENTED = "agent_restful_user_implemented"
    WEBSOCKET_USER_IMPLEMENTED = "agent_websocket_user_implemented"
    ACTION = "agent_action"


class FillerAudioConfig(BaseModel):
    silence_threshold_seconds: float = FILLER_AUDIO_DEFAULT_SILENCE_THRESHOLD_SECONDS
    use_phrases: bool = True
    use_typing_noise: bool = False

    @validator("use_typing_noise")
    def typing_noise_excludes_phrases(cls, v, values):
        if v and values.get("use_phrases"):
            values["use_phrases"] = False
        if not v and not values.get("use_phrases"):
            raise ValueError("must use either typing noise or phrases for filler audio")
        return v


class WebhookConfig(BaseModel):
    url: str


class AzureOpenAIConfig(BaseModel):
    base_url: str
    api_key: str
    region: str
    deployment_name: str
    openai_model_name: str
    api_type: str = AZURE_OPENAI_DEFAULT_API_TYPE
    api_version: Optional[str] = AZURE_OPENAI_DEFAULT_API_VERSION


class CutOffResponse(BaseModel):
    messages: List[BaseMessage] = [BaseMessage(text="Sorry?")]


class AgentConfig(TypedModel, type=AgentType.BASE.value):  # type: ignore
    initial_message: Optional[BaseMessage] = None
    generate_responses: bool = True
    allowed_idle_time_seconds: Optional[float] = None
    num_check_human_present_times: int = 0
    allow_agent_to_be_cut_off: bool = True
    end_conversation_on_goodbye: bool = False
    send_filler_audio: Union[bool, FillerAudioConfig] = False
    webhook_config: Optional[WebhookConfig] = None
    actions: Optional[List[ActionConfig]] = None
    initial_message_delay: float = 0.0
    goodbye_phrases: Optional[List[str]] = None
    interrupt_sensitivity: InterruptSensitivity = "low"
    cut_off_response: Optional[CutOffResponse] = None


class LLMAgentConfig(AgentConfig, type=AgentType.LLM.value):  # type: ignore
    prompt_preamble: str
    model_name: str = LLM_AGENT_DEFAULT_MODEL_NAME
    temperature: float = LLM_AGENT_DEFAULT_TEMPERATURE
    max_tokens: int = LLM_AGENT_DEFAULT_MAX_TOKENS


class LLMFallback(BaseModel):
    provider: Literal["openai", "azure"]
    model_name: str


class ChatGPTAgentConfig(AgentConfig, type=AgentType.CHAT_GPT.value):  # type: ignore
    openai_api_key: Optional[str] = None
    prompt_preamble: str
    model_name: str = CHAT_GPT_AGENT_DEFAULT_MODEL_NAME
    temperature: float = LLM_AGENT_DEFAULT_TEMPERATURE
    max_tokens: int = LLM_AGENT_DEFAULT_MAX_TOKENS
    azure_params: Optional[AzureOpenAIConfig] = None
    vector_db_config: Optional[VectorDBConfig] = None
    # TODO: the below fields should moved up to AgentConfig, and their logic should live in BaseAgent
    use_backchannels: bool = False
    backchannel_probability: float = 0.7
    first_response_filler_message: Optional[str] = None
    llm_fallback: Optional[LLMFallback] = None


class AnthropicAgentConfig(AgentConfig, type=AgentType.ANTHROPIC.value):  # type: ignore
    prompt_preamble: str
    model_name: str = CHAT_ANTHROPIC_DEFAULT_MODEL_NAME
    max_tokens: int = LLM_AGENT_DEFAULT_MAX_TOKENS
    temperature: float = LLM_AGENT_DEFAULT_TEMPERATURE


class ChatVertexAIAgentConfig(AgentConfig, type=AgentType.CHAT_VERTEX_AI.value):  # type: ignore
    prompt_preamble: str
    model_name: str = CHAT_VERTEX_AI_DEFAULT_MODEL_NAME
    generate_responses: bool = False  # Google Vertex AI doesn't support streaming


class InformationRetrievalAgentConfig(
    AgentConfig, type=AgentType.INFORMATION_RETRIEVAL.value  # type: ignore
):
    recipient_descriptor: str
    caller_descriptor: str
    goal_description: str
    fields: List[str]
    # TODO: add fields for IVR, voicemail


class EchoAgentConfig(AgentConfig, type=AgentType.ECHO.value):  # type: ignore
    pass


class GPT4AllAgentConfig(AgentConfig, type=AgentType.GPT4ALL.value):  # type: ignore
    prompt_preamble: str
    model_path: str
    generate_responses: bool = False


class RESTfulUserImplementedAgentConfig(
    AgentConfig, type=AgentType.RESTFUL_USER_IMPLEMENTED.value  # type: ignore
):
    class EndpointConfig(BaseModel):
        url: str
        method: str = "POST"

    respond: EndpointConfig
    generate_responses: bool = False
    # generate_response: Optional[EndpointConfig]


class RESTfulAgentInput(BaseModel):
    conversation_id: str
    human_input: str


class RESTfulAgentOutputType(str, Enum):
    BASE = "restful_agent_base"
    TEXT = "restful_agent_text"
    END = "restful_agent_end"


class RESTfulAgentOutput(TypedModel, type=RESTfulAgentOutputType.BASE):  # type: ignore
    pass


class RESTfulAgentText(RESTfulAgentOutput, type=RESTfulAgentOutputType.TEXT):  # type: ignore
    response: str


class RESTfulAgentEnd(RESTfulAgentOutput, type=RESTfulAgentOutputType.END):  # type: ignore
    pass



import asyncio
from typing import Literal

import aiohttp
from aiohttp import ClientTimeout

from vocode.streaming.models.audio import AudioEncoding
from vocode.streaming.models.message import BaseMessage
from vocode.streaming.models.synthesizer import PlayHtSynthesizerConfig
from vocode.streaming.synthesizer.base_synthesizer import BaseSynthesizer, SynthesisResult

TTS_ENDPOINT = "https://play.ht/api/v2/tts/stream"


class PlayHtSynthesizer(BaseSynthesizer[PlayHtSynthesizerConfig]):
    def __init__(
        self,
        synthesizer_config: PlayHtSynthesizerConfig,
        max_backoff_retries=3,
        backoff_retry_delay=2,
    ):
        super().__init__(
            synthesizer_config,
        )
        self.synthesizer_config = synthesizer_config
        self.api_key = synthesizer_config.api_key
        self.user_id = synthesizer_config.user_id
        if not self.api_key or not self.user_id:
            raise ValueError(
                "You must set the PLAY_HT_API_KEY and PLAY_HT_USER_ID environment variables"
            )
        self.words_per_minute = 150
        self.experimental_streaming = synthesizer_config.experimental_streaming
        self.max_backoff_retries = max_backoff_retries
        self.backoff_retry_delay = backoff_retry_delay

    async def create_speech_uncached(
        self,
        message: BaseMessage,
        chunk_size: int,
        is_first_text_chunk: bool = False,
        is_sole_text_chunk: bool = False,
    ) -> SynthesisResult:
        self.total_chars += len(message.text)
        headers = {
            "AUTHORIZATION": f"Bearer {self.api_key}",
            "X-USER-ID": self.user_id,
            "Content-Type": "application/json",
        }
        output_format: Literal["wav", "mulaw"]
        if self.synthesizer_config.audio_encoding == AudioEncoding.LINEAR16:
            output_format = "wav"
        elif self.synthesizer_config.audio_encoding == AudioEncoding.MULAW:
            output_format = "mulaw"
        body = {
            "quality": "draft",
            "voice": self.synthesizer_config.voice_id,
            "text": message.text,
            "sample_rate": self.synthesizer_config.sampling_rate,
            "output_format": output_format,
        }
        if self.synthesizer_config.speed:
            body["speed"] = self.synthesizer_config.speed
        if self.synthesizer_config.seed:
            body["seed"] = self.synthesizer_config.seed
        if self.synthesizer_config.temperature:
            body["temperature"] = self.synthesizer_config.temperature
        if self.synthesizer_config.quality:
            body["quality"] = self.synthesizer_config.quality

        backoff_retry_delay = self.backoff_retry_delay
        max_backoff_retries = self.max_backoff_retries
        for attempt in range(max_backoff_retries):
            response = await self.async_requestor.get_session().post(
                TTS_ENDPOINT,
                headers=headers,
                json=body,
                timeout=ClientTimeout(total=15),
            )
            if not response.ok:
                raise Exception(f"Play.ht API error status code {response.status}")
            if response.status == 429 and attempt < max_backoff_retries - 1:
                await asyncio.sleep(backoff_retry_delay)
                backoff_retry_delay *= 2  # Exponentially increase delay
                continue

            if not response.ok:
                raise Exception(f"Play.ht API error status code {response.status}")

            if self.experimental_streaming:
                return SynthesisResult(
                    self.experimental_mp3_streaming_output_generator(
                        response, chunk_size
                    ),  # should be wav
                    lambda seconds: self.get_message_cutoff_from_voice_speed(
                        message, seconds, self.words_per_minute
                    ),
                )
            else:
                return SynthesisResult(
                    self._streaming_chunk_generator(response, chunk_size, output_format),
                    lambda seconds: self.get_message_cutoff_from_voice_speed(message, seconds, 150),
                )

        raise Exception("Max retries reached for Play.ht API")

    @classmethod
    def get_voice_identifier(cls, synthesizer_config: PlayHtSynthesizerConfig):
        return ":".join(
            (
                "play_ht",
                str(synthesizer_config.user_id),
                synthesizer_config.voice_id,
                str(synthesizer_config.speed),
                str(synthesizer_config.seed),
                str(synthesizer_config.temperature),
                str(synthesizer_config.quality),
                synthesizer_config.audio_encoding,
            )
        )

    @staticmethod
    async def _streaming_chunk_generator(
        response: aiohttp.ClientResponse,
        chunk_size: int,
        output_format: Literal["wav", "mulaw"],
    ):
        if output_format == "wav":
            buffer = b""
            is_first_chunk = True
            async for chunk in response.content.iter_any():
                if is_first_chunk:
                    is_first_chunk = False
                    buffer += chunk[88:]  # size of the wav header
                else:
                    buffer += chunk
                i = 0
                while i < len(buffer) - chunk_size:
                    yield SynthesisResult.ChunkResult(buffer[i : i + chunk_size], False)
                    i += chunk_size
                buffer = buffer[i:]
        elif output_format == "mulaw":
            async for chunk in response.content.iter_chunked(chunk_size):
                yield SynthesisResult.ChunkResult(chunk, False)
        await response.release()



import os
from contextvars import ContextVar, Token
from typing import Any
from uuid import UUID

import sentry_sdk
from loguru import logger

from vocode.meta import ensure_punkt_installed

environment = {}
logger.disable("vocode")

ensure_punkt_installed()


class ContextWrapper:
    """Context Variable Wrapper."""

    _instances: list = []

    def __init__(self, value: ContextVar) -> None:
        self.__value: ContextVar = value
        self.__token: Token
        ContextWrapper._instances.append(self)

    def set(self, value: Any) -> Token:
        """Set a context variable."""
        self.__token = self.__value.set(value)
        if isinstance(value, str):
            sentry_sdk.set_tag(self.__value.name, value)
        if isinstance(value, UUID):
            sentry_sdk.set_tag(self.__value.name, str(value))

        return self.__token

    def reset(self, token: Token | None = None) -> None:
        """Reset a context variable."""
        if not hasattr(self, "__token"):
            return

        if not token:
            token = self.__token
        self.__value.reset(token)
        return

    def __module__(self) -> Any:  # type: ignore
        return self.__value.get()

    @property
    def value(self) -> Any:
        """Gets the value of a context variable."""
        return self.__value.get()

    @classmethod
    def serialize_instances(cls) -> dict:
        """Gathers all instances of ContextWrapper."""
        instances = {}
        for instance in cls._instances:
            value = instance.__value.get()
            if isinstance(value, UUID):
                value = str(value)

            if isinstance(value, str):
                instances[instance.__value.name] = value
        return instances


def setenv(**kwargs):
    for key, value in kwargs.items():
        environment[key] = value


def getenv(key, default=None):
    return environment.get(key) or os.getenv(key, default)


api_key = getenv("VOCODE_API_KEY")
base_url = getenv("VOCODE_BASE_URL", "api.vocode.dev")


conversation_id: ContextWrapper = ContextWrapper(
    ContextVar("conversation_id", default=None),
)
sentry_span_tags: ContextWrapper = ContextWrapper(ContextVar("sentry_span_tags", default=None))
get_serialized_ctx_wrappers = ContextWrapper.serialize_instances
